#include <cstdio>
#include <iostream>
#include <vector>
#include <queue>
#include <iomanip>
#include <cstdlib>
#include <ctime>
#include <unistd.h>
#include <algorithm>

using namespace std;

#include "common.h"
#include "marshal.h"
#include "update.h"

#define VEZA_NONE -1

struct policko {
	int x;
	int y;
	int capacity;	
	int type;
	double defenseImp;
	double enemyrate;
	double enemykill;
	
	Veza tower;
	
	policko (int i, int j) {
		x = i;
		y = j;
	}
	
	policko (){
		x = -1;
		y = -1;
		enemyrate = 0;
		defenseImp = 0;
		enemykill = 0;
	}
};


Mapa mapa;
Stav stav;   
vector<Prikaz> prikazy;
vector <vector<policko> > map;
vector <policko> enemyPath;
vector <policko> defensePositions;


bool vykonaj(const Prikaz& p) {
	bool uspech = vykonajPrikaz(mapa, stav, 0, p);
	if (uspech) prikazy.push_back(p);   
	return uspech;
}


void computeNodes() {						
	for (int i=0; i<mapa.w; i++){
		for (int j=0; j<mapa.h; j++){
			policko temp(i,j);
			temp.capacity = mapa.loziska[j][i];
			temp.type = mapa.zisti(i, j);
			map[i][j] = temp;
		}
	}
}

double dist (int x1, int y1, int x2, int y2){
	return (x2-x1)*(x2-x1) + (y2-y1)*(y2-y1);
}


void refreshTowers (){
	for (int i=0; i<mapa.w; i++){
		for (int j=0; j<mapa.h; j++){
			map[i][j].tower.typ = VEZA_NONE;
		}
	}

	
	for (int i=0; i<stav.hraci[0].veze.size(); i++){
		Veza temp = stav.hraci[0].veze[i];
		map[temp.x][temp.y].tower = temp;		
	}
}

int dir [][2] = {{0, 1}, {1, 0}, {0, -1}, {-1, 0}};

int enemyPathBfs (int x, int y, vector<vector<int> > & bfsdata) {
	
	int res = 0;
	
	queue <policko> Q;
	bfsdata[x][y] = 0;
	
	Q.push(policko (x, y));
	Q.push(policko());
	
	int count = 1;
	
	while (! Q.empty()){
		policko temp = Q.front();
		Q.pop();
		
		if (temp.x == -1) {
			count += 1;
			if (Q.empty()) return res;
			Q.push(policko());
		}
		else {
			for (int i=0; i<4; i++) {
				int x1 = temp.x + dir[i][0];
				int y1 = temp.y + dir[i][1];
				
				if ((mapa.priechodne(x1, y1)) && (bfsdata[x1][y1] == -1)){
					Q.push(policko(x1,y1));
					bfsdata[x1][y1] = count;
					if (mapa.zisti(x1, y1) == MAPA_CIEL) res = count;
				}
			}
		}
	}
		
	return res;
}


double enemyPathDfs (int x, int y, int maxcount, const vector<vector<int> > &bfsdata, double enemyrate = 100, int count=0, bool test = false) {
	if (!mapa.priechodne(x, y)) return 0;
	if (count > maxcount) return 0;
	
	if (mapa.zisti(x, y) == MAPA_CIEL) return enemyrate;
	
	int neibcount = 0;
	
	for (int i=0; i<4; i++) {
		int x1 = x + dir[i][0];
		int y1 = y + dir[i][1];
		
		if (!mapa.priechodne(x1, y1)) continue;
		if (bfsdata[x1][y1] <= count) continue;
		if (enemyPathDfs (x1, y1, maxcount, bfsdata, 100, count + 1, true) > 0) neibcount++;
	}
	
	if (neibcount == 0) return 0;
	if (test) return enemyrate;
	
	
	for (int i=0; i<4; i++) {
		int x1 = x + dir[i][0];
		int y1 = y + dir[i][1];
		
		if (!mapa.priechodne(x1, y1)) continue;
		if (bfsdata[x1][y1] <= count) continue;
		enemyPathDfs (x1, y1, maxcount, bfsdata, (enemyrate) / neibcount , count + 1);
	}	
	
	if (count == 0) return enemyrate;
	
	map[x][y].enemyrate += enemyrate;
	return enemyrate;
}

bool CompByDefImp (policko a, policko b) {
	return a.defenseImp > b.defenseImp;
}


void computeEnemyPath () {
	vector<policko> spawns;
	
	for (int i=0; i<mapa.w; i++){
		for (int j=0; j<mapa.h; j++){
			if (mapa.zisti(i, j) == MAPA_SPAWN) spawns.push_back(policko(i, j));
			map[i][j].enemyrate = 0;
		}
	}
	
	vector <vector<int> > bfsdata;
	for (int i=0; i<spawns.size(); i++){
		bfsdata = vector<vector<int> > (mapa.w, vector <int> (mapa.h, -1));
		int distance = enemyPathBfs (spawns[i].x, spawns[i].y, bfsdata);
		enemyPathDfs (spawns[i].x, spawns[i].y, distance, bfsdata);		
	}
	
/*	for (int i=0; i<mapa.h; i++){
		for (int j=0; j<mapa.w; j++){
			cerr << setw(5)<< map[j][i].enemyrate << " ";
		}
		cerr << endl;
	}
	cerr << endl;
	
	
	for (int i=0; i<mapa.h; i++){
		for (int j=0; j<mapa.w; j++){
			cerr << setw(5)<< map[j][i].enemykill << " ";
		}
		cerr << endl;
	}
	cerr << endl;*/
}


double computeDefenseImp (int x, int y){
	double res = 0;
	
	if (mapa.zisti(x, y) != MAPA_POZEMOK) return 0;
	
	for (int i=0; i<4; i++) {
		int x1 = x + dir[i][0];
		int y1 = y + dir[i][1];
		
		if (!mapa.priechodne(x1, y1)) continue;
		res += map[x1][y1].enemyrate;
	}
	
	return res;
}


void computeDefenseImpNodes () {
	for (int i=0; i<mapa.w; i++){
		for (int j=0; j<mapa.h; j++){
			map[i][j].defenseImp = computeDefenseImp(i, j);
		}
	}
	
	/*for (int i=0; i<mapa.h; i++){
		for (int j=0; j<mapa.w; j++){
			cerr << setw(5)<< map[j][i].defenseImp << " ";
		}
		cerr << endl;
	}
	cerr << endl;*/
	
	defensePositions.clear();
	
	for (int i=0; i<mapa.w; i++){
		for (int j=0; j<mapa.h; j++){
			if (map[i][j].defenseImp < 3) continue;
			defensePositions.push_back(map[i][j]);
		}
	}
	
	sort (defensePositions.begin(), defensePositions.end(), CompByDefImp);
		
	//cerr << endl;
}


int akoBudemUtocit;
int attack;

void inicializuj() {
	srand(time(NULL) * getpid() + 42);
	akoBudemUtocit = rand() % UTOCNIK_POCET_TYPOV;	
	attack = 1;
	
	cerr.precision(0);
	cerr.setf(ios::fixed, ios::floatfield);

	map = vector<vector<policko> > (mapa.w, vector<policko> (mapa.h, policko()));
	computeNodes();
	computeEnemyPath();
	computeDefenseImpNodes();
}

int pocetVezi(int typ) {
	int result = 0;
	FOREACH(it, stav.hraci[0].veze) if (it->typ == typ) result++;
	return result;
}

int countAtk (){
	int res = 0;
	for (int i=0; i<VEZA_POCET_BOJOVYCH; i++){
		res+= pocetVezi(i);
	}
	return res;
}



void zistiTah() {
	

	refreshTowers();
	computeEnemyPath();
	computeDefenseImpNodes();
	
	int mojLab = VEZA_LAB_MUCHA;
		
	/*while (pocetVezi(mojLab) < 1) {
		int x, y;
		do {
			x = rand() % mapa.w;
			y = rand() % mapa.h;
		}
		while(map[x][y].defenseImp > 5);
			
		if (!vykonaj(Prikaz::buduj(x, y, mojLab))) break;
	}
	
	if (attack != 0) {			
		int naKoho = 1 + rand() % (mapa.pocetHracov-1);
		vykonaj(Prikaz::utoc(mojLab-VEZA_LAB_PRVY, naKoho));
		attack--;
	}*/
	
	
	
	for (int i=0; i<50; i++)
		while ((pocetVezi(VEZA_ZBERAC) < 1)||(double)countAtk() / pocetVezi(VEZA_ZBERAC)+1 > 2) {
			int x, y;
			do {
				x = rand() % mapa.w;
				y = rand() % mapa.h;
			}
			while(map[x][y].defenseImp > 5);
			
			if (!vykonaj(Prikaz::buduj(x, y, VEZA_ZBERAC))) break;
		}
	
	for (int i=0; i<defensePositions.size(); i++) {
		int x = defensePositions[i].x;
		int y = defensePositions[i].y;
		
		if (x == -1) continue;
			
		for (int i=0; i<10; i++) {
			int type = rand() % 7; 
			if (!vykonaj(Prikaz::buduj(x, y, type))) break;
		}
	}
}

int main() {

	nacitaj(cin, mapa);
	inicializuj();

	while (cin.good()) {
		nacitaj(cin, stav);
		prikazy.clear();
		zistiTah();
		uloz(cout, prikazy);
		cout << ".\n" << flush;
	}

	return 0;
}

