
from collections import namedtuple


def file_source(filename):
    with open(filename) as fh:
        for line in fh:
            for number in line.split():
                yield int(number)


def string_source(string):
    return iter(map(int, string.split()))


class _int(object):
    @staticmethod
    def load(source):
        try:
            return next(source)
        except StopIteration:
            raise EOFError()


class vector(object):
    def __init__(self, T):
        self.T = T

    def load(self, source):
        size = _int.load(source)
        return [self.T.load(source) for i in xrange(size)]


class struct(object):
    @classmethod
    def load(cls, source):
        self = cls()
        for fieldtype, fieldname in cls._fields:
            setattr(self, fieldname, fieldtype.load(source))
        return self


def load_log(source):
    stavy = [Stav.load(source)]
    while True:
        try:
            vector(Odpoved).load(source)
            stavy.append(Stav.load(source))
        except EOFError:
            break
    return stavy


def load_observation(filename):
    result = []
    with open(filename) as fh:
        for line in fh:
            line = line.split()
            result.append(line[0:1] + map(int, line[1:]))
    return result


class Prikaz(struct):
    _fields = [
        (_int, 'typ'),
        (_int, 'x'),
        (_int, 'y'),
        (_int, 'a'),
        (_int, 'b'),
    ]

Odpoved = vector(Prikaz)

class Veza(struct):
    _fields = [
        (_int, 'x'),
        (_int, 'y'),
        (_int, 'typ'),
        (_int, 'energia'),
        (_int, 'baseLevel'),
        (vector(_int), 'levelBonusy'),
        (_int, 'terazTahala'),
    ]

class Utocnik(struct):
    _fields = [
        (_int, 'x'),
        (_int, 'y'),
        (_int, 'typ'),
        (_int, 'hp'),
        (_int, 'ktorehoHraca'),
        (_int, 'pohybovyTimer'),
        (_int, 'kolkoSpomaleny'),
    ]

class Hrac(struct):
    _fields = [
        (_int, 'body'),
        (_int, 'drevo'),
        (_int, 'umrel'),
        (vector(Veza), 'veze'),
        (vector(Utocnik), 'utocnici'),
        (vector(Utocnik), 'prichadzajuci'),
        (vector(_int), 'mapovanie'),
    ]

class Stav(struct):
    _fields = [
        (vector(Hrac), 'hraci'),
        (_int, 'cas'),
    ]

class Mapa(struct):
    _fields = [
        (_int, 'pocetHracov'),
        (_int, 'w'),
        (_int, 'h'),
        (vector(vector(_int)), 'pole'),
        (vector(vector(_int)), 'loziska'),
    ]


def set_format_version(version):
    if version >= 2:
        Utocnik._fields.insert(
            Utocnik._fields.index((_int, 'pohybovyTimer')),
            (_int, 'level'))

